/*
 * field_model.c - functions to predict the earth's geomagnetic field
 *
 * Copyright (C) 1998, John C. Peterson <mailto:jaypee@netcom.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2. A copy is included in this distribution in the file
 * named "LICENSE".
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License, version 2, for more details.
 *
 * $Id: field_model.c,v 1.1 1998/05/05 05:59:16 jcp Exp $
 *
 */

#include <stdio.h>
#include <math.h>

#include "field_model.h"

/* define macros */

#define MAX_DEGREE	16
#define MAX_TERMS	(((MAX_DEGREE+3)*(MAX_DEGREE))/2)
#define MAX_FACTORIAL	(2*(MAX_DEGREE+1))

#define RAD_EARTH	(6371.2L)	/* mean radius of IAU-66 ellipsoid */
#define RAD_MAJOR	(6378.137L)	/* major axis of WGS-84 ellipsoid */
#define RAD_MINOR	(6356.7523142L)	/* minor axis of WGS-84 ellipsoid */
#define ECCENTRICITY2	(0.00669437999013L)	/* eccentricity^2 for WGS-84 */

#define SQR(x)		((x)*(x))
#define RAD2DEG(rad)	((rad)*(180.0L/M_PI))
#define DEG2RAD(deg)	((deg)*(M_PI/180.0L))

#define TRUE		((int)1)
#define FALSE		((int)0)

/* character buffers */

#define ERROR_LEN	256

static char err_str[ERROR_LEN];

#define LINE_LEN	256

static char line_buf[LINE_LEN];

/* model description */

static int initialized = FALSE;

static char model_name[LINE_LEN];
static int model_degree;
static int secular_degree;
static double base_epoch;
static double min_date;
static double max_date;
static double min_altitude;
static double max_altitude;

/* model coefficients */

static int num_terms;
static int degree[MAX_TERMS];
static int order[MAX_TERMS];
static double g[MAX_TERMS];
static double h[MAX_TERMS];
static double dgdt[MAX_TERMS];
static double dhdt[MAX_TERMS];

/* -------------------------------------------------------------------------- */

extern char *strerror_field_model(void)
{
  return err_str;
}

/* -------------------------------------------------------------------------- */

extern int init_field_model(char *filename)
{

  FILE *fpmodel;

  char order_str[3];
  char degree_str[3];
  char g_str[9];
  char h_str[9];
  char dgdt_str[9];
  char dhdt_str[9];

  int nscan;
  int i, m, n;

  double factorial[MAX_FACTORIAL];
  double scale;

  /* attempt to open model data file */

  if ((fpmodel = fopen(filename, "r")) == NULL) {
    sprintf(err_str, "init_field_model: open of model data file failed: %80s\n",
                     filename);
    return FALSE;
  }

  /*
  **
  ** read the header line from the model data file
  **
  **   model name
  **   base epoch
  **   max degree, order of field model
  **   max degree, order of secular change model
  **   max degree, order of acceleration model (value ignored by this function)
  **   min, max valid dates (year of gregorian calendar)
  **   min, max valid altitudes (kilometers)
  **
  */

  if (fgets(line_buf, sizeof(line_buf), fpmodel) != NULL) {

    nscan = sscanf(line_buf, "%s%lf%d%d%*d%lf%lf%lf%lf",
                             model_name,
                             &base_epoch,
                             &model_degree,
                             &secular_degree,
                             &min_date,
                             &max_date,
                             &min_altitude,
                             &max_altitude);

    if (nscan < 7) {
      /* mangled header line */
      sprintf(err_str,
              "init_field_model: model data file has mangled header: %80s\n",
              line_buf);
      fclose(fpmodel);
      return FALSE;
    }

  }

  /* verify the degree of the field model is not too large */

  if (model_degree > MAX_DEGREE) {
    sprintf(err_str,
            "init_field_model: model degree exceeds internal storage: %d\n",
            model_degree);
    return FALSE;
  }

  /* initialize the spherical harmonic coefficients to zero */

  for (i = 0; i < MAX_TERMS; i++) {
    g[i] = dgdt[i] = 0.0L;
    h[i] = dhdt[i] = 0.0L;
  }

  /* read the spherical harmonic coefficients from the model data file */

  i = 0;

  order_str[2] = '\0';
  degree_str[2] = '\0';
  g_str[8] = '\0';
  h_str[8] = '\0';
  dgdt_str[8] = '\0';
  dhdt_str[8] = '\0';

  while (fgets(line_buf, sizeof(line_buf), fpmodel) != NULL) {

    nscan = sscanf(line_buf, "%2c%2c%8c%8c%8c%8c",
                             order_str,
                             degree_str,
                             g_str,
                             h_str,
                             dgdt_str,
                             dhdt_str);

    if (nscan == 2) {
      if ((strcmp(order_str,"99") == 0) && (strcmp(degree_str,"99") == 0)) {
        /* explicit end of coefficients marker */
        break;
      }
    }

    if (nscan == 6) {
      if (i < MAX_TERMS) {
        nscan = 0;
        /* convert the character strings */
        nscan += sscanf(order_str, "%d", &order[i]);
        nscan += sscanf(degree_str, "%d", &degree[i]);
        nscan += sscanf(g_str, "%lf", &g[i]);
        nscan += sscanf(h_str, "%lf", &h[i]);
        nscan += sscanf(dgdt_str, "%lf", &dgdt[i]);
        nscan += sscanf(dhdt_str, "%lf", &dhdt[i]);
        if (nscan < 6) {
          sprintf(err_str,
                  "init_field_model: error parsing coefficients line: %80s\n",
                  line_buf);
          fclose(fpmodel);
          return FALSE;
        }
      } else {
        sprintf(err_str,
                "init_field_model: read unexpected coefficients line: %80s\n",
                line_buf);
        fclose(fpmodel);
        return FALSE;
      }
    } else {
      sprintf(err_str,
              "init_field_model: error parsing coefficients line: %80s\n",
              line_buf);
      fclose(fpmodel);
      return FALSE;
    }

    i++;

  }

  num_terms = i;

  fclose(fpmodel);

  /*
  ** If no secular change model was provided, (flagged by specifying
  ** zero for the degree of the secular change model on the header
  ** line), then zero out all the secular change model coefficients
  ** again to be on the safe side.
  */

  if (secular_degree == 0) {
    for (i = 0; i < num_terms; i++) {
      dgdt[i] = 0.0L;
      dhdt[i] = 0.0L;
    }
  }

  /*
  ** The models use Schmidt quasi-normalized associated Legendre functions.
  ** It is easiest for us to just compute the "vanilla" associated Legendre
  ** functions (which use the normalization factor K_n = 1 / 2^n), when we
  ** evaluate the series. We multiply the model coefficients by the correct
  ** scale factors that relate the two associated Legendre functions, or
  ** sqrt( 2 * (n-m)! / (n+m)! ) for m > 0, to be specific.
  */

  factorial[0] = 1.0L;

  for (n = 1; n <= 2*model_degree; n++) {
    factorial[n] = factorial[n-1] * n;
  }

  for (i = 0; i < num_terms; i++) {
    n = degree[i];
    m = order[i];
    scale = (m > 0) ? sqrt(2.0L * factorial[n-m] / factorial[n+m]) : 1.0L;
    g[i] *= scale;
    h[i] *= scale;
    dgdt[i] *= scale;
    dhdt[i] *= scale;
  }

  /* normal exit */

  initialized = TRUE;

  return TRUE;

}

/* -------------------------------------------------------------------------- */

extern int field_model(field_model_t *value)
{

  int i, m, n;

  double g_t;
  double h_t;
  double rho;
  double cos_phi;
  double sin_phi;
  double cos_lat;
  double sin_lat;
  double cos_theta;
  double sin_theta;
  double x_p, z_p, rc;
  double cos_alpha;
  double sin_alpha;
  double Br, Bp, Bt;
  double Bn, Be, Bh, Bz;
  double a, b;

  double r_n[MAX_DEGREE+1];
  double cos_mphi[MAX_DEGREE+1];
  double sin_mphi[MAX_DEGREE+1];
  double P[MAX_DEGREE+1][MAX_DEGREE+1];		/* P[degree=n][order=m]; */
  double dPdx[MAX_DEGREE+1][MAX_DEGREE+1];	/* dPdx[degree=n][order=m]; */

  /* check that the model was properly initialized */

  if (! initialized) {
    sprintf(err_str, "field_model: model was not initialized\n");
    return FALSE;
  }

  /* check for valid latitude */

  if (fabs(value->latitude) > 80.0) {
    sprintf(err_str, "field_model: input latitude is out of range: %g\n",
                     value->latitude);
    return FALSE;
  }

  /* check for valid longitude */

  if (fabs(value->longitude) > 180.0) {
    sprintf(err_str, "field_model: input longitude is out of range: %g\n",
                     value->longitude);
    return FALSE;
  }

  /* check for valid altitude */

  if ((value->altitude < min_altitude) || (value->altitude > max_altitude)) {
    sprintf(err_str, "field_model: input altitude is out of range: %g\n",
                     value->altitude);
    return FALSE;
  }

  /* check for valid date */

  if ((value->date < min_date) || (value->date > max_date)) {
    sprintf(err_str, "field_model: input date is out of range: %g\n",
                     value->date);
    return FALSE;
  }

  /*
  ** Now we convert from geodetic coordinates, (defined by the WGS-84
  ** reference ellipsoid), to Earth Centered Earth Fixed Cartesian
  ** coordinates, and then to spherical coordinates. 
  */

  cos_phi = cos(DEG2RAD(value->longitude));
  sin_phi = sin(DEG2RAD(value->longitude));

  cos_lat = cos(DEG2RAD(value->latitude));
  sin_lat = sin(DEG2RAD(value->latitude));

  /* compute the local radius of curvature on the WGS-84 reference ellipsoid */

  rc = RAD_MAJOR / sqrt(1.0L - ECCENTRICITY2*SQR(sin_lat));

  /* compute ECEF cartesian coordinates of specified point (for longitude=0) */

  x_p = (rc + value->altitude) * cos_lat;
  z_p = (rc*(1.0L - ECCENTRICITY2) + value->altitude) * sin_lat;

  /* compute spherical radius and angle theta of specified point */

  rho = sqrt(SQR(x_p) + SQR(z_p));
  cos_theta = z_p / rho;
  sin_theta = x_p / rho;

  /* compute the angle between the spherical and WGS-84 basis vectors */

  sin_alpha = cos_lat*cos_theta - sin_lat*sin_theta;
  cos_alpha = sqrt((1.0L + sin_alpha)*(1.0L - sin_alpha));

  /* for n = 0 ... model_order, compute (RAD_EARTH/rho)^(n+2) */

  r_n[0] = SQR(RAD_EARTH / rho);

  for (n = 1; n <= model_degree; n++) {
    r_n[n] = r_n[n-1] * (RAD_EARTH / rho);
  }

  /*
  ** Compute cos(m*phi), sin(m*phi) for m = 0 ... model_degree
  ** by recursion using the familiar angle sum identidy:
  **   cos(a + b) = cos(a)*cos(b) - sin(a)*sin(b)
  **   sin(a + b) = cos(a)*sin(b) + sin(a)*cos(b)
  */

  cos_mphi[0] = 1.0L;
  sin_mphi[0] = 0.0L;

  cos_mphi[1] = cos_phi;
  sin_mphi[1] = sin_phi;

  for (m = 2; m <= model_degree; m++) {
    cos_mphi[m] = cos_mphi[m-1]*cos_phi - sin_mphi[m-1]*sin_phi;
    sin_mphi[m] = cos_mphi[m-1]*sin_phi + sin_mphi[m-1]*cos_phi;
  }

  /*
  ** Compute the associated Legendre function, and it's derivative for
  ** degrees n = 1 ... model_degree, and orders m = 0 ... n for the argument
  ** cos( theta = latitude ). For a derivation of the algorithm used below,
  ** see G.R. Gathers, "Routines using Associated Legendre Functions and
  ** Legendre Polynomials", University of California - Lawrence Livermore
  ** Lab report, Jan 31, 1977, UCRL-52127. An online version in pdf format        ** is available from;  http://www.llnl.gov/tid/lof/
  */

  /*
  ** Begin by computing P[n][m] for n = 1 ... model_degree, and m = n.
  ** The values are found by direct (but tedious) evaluation of Rodriguez
  ** Formula to give: P[n][m](x) = (1 - x^2)^(n/2) * (2n)! / ( 2^n * n! ).
  ** For simplicity, note that (2n)! / n! = (2n)*(2n-1)* ... *(n+2)*(n+1),
  ** the later being easier to update in a recursion based computation.
  ** Note also that since x = cos(theta), (1 - x^2)^(1/2) = sin(theta).
  */

  P[0][0] = 1.0L;

  for (n = 1; n <= model_degree; n++) {
    P[n][n] = sin_theta * P[n-1][n-1] * (2*n - 1);
  }

  /*
  ** Next, we compute P[n][m] for n = 1 ... model_degree, and m = n-1.
  ** Again, the values are found by direct evaluation of Rodriguez
  ** Formula: P[n][m](x) = x * (1 - x^2)^(m/2) * (2n)! / ( 2^n * n! )
  ** which can be written as: ( 2*m + 1 ) * x * P[m][m] for use in
  ** a recursive based computation.
  */

  for (n = 1; n <= model_degree; n++) {
    m = n - 1;
    P[n][m] = (2*m + 1) * cos_theta * P[m][m];
  }

  /*
  ** Next, we compute the remaining associated Legendre functions
  ** P[n][m] for each m = 0 ... model_degree, and for each n = m+2
  ** ... model_degree, using forward recursion in order n.
  */

  for (m = 0; m <= model_degree; m++) {
    for (n = m+2; n <= model_degree; n++) {
      a = 2*n - 1;
      b = n + m - 1;
      P[n][m] = (a*cos_theta*P[n-1][m] - b*P[n-2][m]) / (n - m);
    }
  }

  /*
  ** To compute dPdx[n][m] for n = 1 ... model_degree, and m = n, we
  ** just differentiate the expression used above for P[n][n](x) to obtain:
  ** dPdx[n][m](x) = n * (-x) * (1 - x^2)^((n-1)/2) * (2n)! / ( 2^n * n! ).
  */

  dPdx[1][1] = -cos_theta / sin_theta;

  for (n = 2; n <= model_degree; n++) {
    a = n * (2.0L*n - 1.0L) / (n - 1.0L);
    dPdx[n][n] = sin_theta * dPdx[n-1][n-1] * a;
  }

  /*
  ** Now we compute the remaining derivatives of the associated Legendre
  ** functions, using a well known forward recursion relation in order n.
  */

  for (m = 0; m <= model_degree; m++) {
    for (n = m+1; n <= model_degree; n++) {
      a = n * cos_theta;
      b = (n + m);
      dPdx[n][m] = (a*P[n][m] - b*P[n-1][m]) / -(sin_theta*sin_theta);
    }
  }

  /*
  ** Finally, we compute the spherical coordinate components of the
  ** geomagnetic field. The geomagnetic potential V is represented
  ** in these models by a summation of spherical harmonics:
  **
  **          N   n        (n+1)  m            m           m
  **  V = RE SUM SUM (RE/r)     [g cos(m p) + h sin(m p)] P (cos(t))
  **         n=1 m=0              n            n           n
  **
  ** where RE is the mean radius of the earth (IAU-66 value), and
  ** (r,t,p) are the rho, theta, phi spherical coordinates. The
  ** vector magnetic field is given by -grad V, where the gradient
  ** in spherical coordinates is given by:
  **
  **           dV ^     1 dV ^        1     dV ^
  **  grad V = -- r  +  - -- t  +  -------- -- p
  **           dr       r dt       r sin(t) dp
  */

  Br = 0.0L;
  Bp = 0.0L;
  Bt = 0.0L;

  for (i = 0; i < num_terms; i++) {

    n = degree[i];
    m = order[i];

    /* adjust the coefficients using the secular change model */
    g_t = g[i] + (value->date - base_epoch)*dgdt[i];
    h_t = h[i] + (value->date - base_epoch)*dhdt[i];

    /* update the rho, theta, and phi components of the B field */
    Br += r_n[n] * (g_t*cos_mphi[m] + h_t*sin_mphi[m]) * (n+1) * P[n][m];
    Bt += r_n[n] * (g_t*cos_mphi[m] + h_t*sin_mphi[m]) * dPdx[n][m];
    Bp += r_n[n] * (h_t*cos_mphi[m] - g_t*sin_mphi[m]) * m * P[n][m];

  }

  /* multiply by constants common to all terms */

  Bt *= sin_theta;
  Bp *= -1.0L / sin_theta;

  /* rotate spherical field components back to the local WGS-84 basis */

  Be = Bp;
  Bn = -cos_alpha * Bt + sin_alpha * Br;
  Bz = -sin_alpha * Bt - cos_alpha * Br;

  /* compute declination, inclination and total intensity */

  Bh = sqrt(SQR(Bn) + SQR(Be));

  value->dec = RAD2DEG(atan2(Be,Bn));
  value->dip = RAD2DEG(atan2(Bz,Bh));
  value->intensity = sqrt(SQR(Bh) + SQR(Bz));

  /* normal exit */

  return TRUE;

}

/* -------------------------------------------------------------------------- */
